;;; -*- syntax: common-lisp; package: clm; base: 10; mode: lisp -*-

(in-package :clm)

;;; foreign function interfaces for ACL, CMUCL, SBCL, OpenMCL


;;; TODO: change all the sound-* to mus-sound-* and deprecate the old forms of
;;;   	  sound-duration sound-chans sound-comment sound-data-format sound-data-location sound-datum-size sound-maxamp
;;;	  sound-format-name sound-header-type sound-length sound-samples sound-frames sound-srate sound-type-name
;;;	  sound-loop-info sound-set-loop-info


(defvar *output* nil)
(defvar *reverb* nil)

(defun clm-close-output () (clm-close-output-1) (setf *output* nil))
(defun clm-close-reverb () (clm-close-reverb-1) (setf *reverb* nil))

(defun sl-dac (name &optional (dev 0)) (sl-dac-1 name dev)) ; for backwards compatibility

#+excl
(progn
  (defun fullname (n)
    (if (stringp n)
	(namestring (full-merge-pathnames n))
      (if (mus-input? n)
	  (mus-file-name n)
	(progn
	  (warn "~A should be a filename" n)
	  n))))
  #-acl-60 (defun sound-comment (arg) (let ((ptr (sound-comment-1 (fullname arg)))) (ff:char*-to-string ptr)))
  #-acl-60 (defun sound-format-name (arg) (ff:char*-to-string (sound-format-name-1 arg)))
  #-acl-60 (defun sound-type-name (arg) (ff:char*-to-string (sound-type-name-1 arg)))
  (defun sound-chans (name) (sound-chans-1 (fullname name)))
  (defun sound-duration (name) (sound-duration-1 (fullname name)))
  (defun sound-data-format (name) (sound-data-format-1 (fullname name)))
  (defun sound-data-location (name) (sound-data-location-1 (fullname name)))
  (defun sound-datum-size (name) (sound-datum-size-1 (fullname name)))
  (defun sound-header-type (name) (sound-header-type-1 (fullname name)))
  (defun sound-length (name) (sound-length-1 (fullname name)))
  (defun sound-samples (name) (sound-samples-1 (fullname name)))
  (defun sound-frames (name) (sound-frames-1 (fullname name)))
  (defun sound-srate (name) (sound-srate-1 (fullname name))))


;;; ---------------------------------------- ACL 4.n ----------------------------------------
#+(and excl (not acl-50))
  (ff:defforeign-list 
   `(
     (acl43-set-arr :entry-point "acl43_set_arr" :arguments (array integer integer integer integer integer) :return-type :void)
     (acl43-get-arr :entry-point "acl43_get_arr" :arguments (array integer array) :return-type :void)
     (mus-set-raw-header-defaults :entry-point "mus_header_set_raw_defaults" :arguments (integer integer integer) :return-type :void)
     (mus-set-header-type :entry-point "mus_file_set_header_type" :arguments (integer integer) :return-type :integer)
     (mus-set-data-clipped :entry-point "mus_file_set_data_clipped" :arguments (integer integer) :return-type :integer)
     (mus-set-chans :entry-point "mus_file_set_chans" :arguments (integer integer) :return-type :integer)
     (array->file :entry-point "clm_array2file" :arguments (string (simple-array double-float (*)) integer integer integer) 
			:return-type :void :prototype t)
     (file->array :entry-point "clm_file2array" :arguments (string integer integer integer (simple-array double-float (*)))
		  :return-type :integer :prototype t)

     (initialize-sndlib :entry-point "mus_sound_initialize" :arguments nil :return-type :integer)
     (initialize-cmus :entry-point "initialize_cmus" :arguments nil :return-type :void)

     #-windoze (clm-send-snd :entry-point "clm_send_snd" :arguments (string) :return-type :integer)
     #-windoze (clm-start-snd :entry-point "clm_start_snd" :arguments nil :return-type :integer)
     #-windoze (clm-init-x :entry-point "clm_init_x" :arguments (string) :return-type :void)
     #-windoze (clm-receive-snd-1 :entry-point "clm_receive_snd" :arguments nil :return-type :integer)

     (sound-chans-1 :entry-point "mus_sound_chans" :arguments (string) :return-type :integer)
     (sound-duration-1 :entry-point "mus_sound_duration" :arguments (string) :return-type :single-float)
     (sound-data-format-1 :entry-point "mus_sound_data_format" :arguments (string) :return-type :integer)
     (sound-data-location-1 :entry-point "clm_sound_data_location" :arguments (string) :return-type :integer)
     (sound-datum-size-1 :entry-point "mus_sound_datum_size" :arguments (string) :return-type :integer)
     (sound-header-type-1 :entry-point "mus_sound_header_type" :arguments (string) :return-type :integer)
     (sound-length-1 :entry-point "clm_sound_length" :arguments (string) :return-type :integer)
     (sound-samples-1 :entry-point "clm_sound_samples" :arguments (string) :return-type :integer)
     (sound-frames-1 :entry-point "clm_sound_frames" :arguments (string) :return-type :integer)
     (sound-srate-1 :entry-point "mus_sound_srate" :arguments (string) :return-type :integer)
     (sound-comment-1 :entry-point "mus_sound_comment" :arguments (string) :return-type :integer)
     (sound-format-name-1 :entry-point "mus_data_format_name" :arguments (integer) :return-type :integer)
     (sound-loop-info :entry-point "clm_sound_loop_info" :arguments (string array) :return-type :integer)
     (sound-set-loop-info :entry-point "mus_sound_set_full_loop_info" :arguments (string array) :return-type :void)
     (sound-type-name-1 :entry-point "mus_header_type_name" :arguments (integer) :return-type :integer)
     (open-sound-input :entry-point "mus_sound_open_input" :arguments (string) :return-type :integer)
     (open-sound-output :entry-point "mus_sound_open_output" :arguments (string integer integer integer integer string) :return-type :integer)
     (reopen-sound-output :entry-point "mus_sound_reopen_output" :arguments (string integer integer integer integer) :return-type :integer)
     (close-sound-input :entry-point "mus_sound_close_input" :arguments (integer) :return-type :integer)
     (close-sound-output :entry-point "mus_sound_close_output" :arguments (integer integer) :return-type :integer)
     (read-sound :entry-point "mus_sound_read" :arguments (integer integer integer integer array) :return-type :integer)
     (write-sound :entry-point "mus_sound_write" :arguments (integer integer integer integer array) :return-type :integer)
     (seek-sound-frame :entry-point "mus_sound_seek_frame" :arguments (integer integer) :return-type :integer)
     (sl-dac-1 :entry-point "sl_dac" :arguments (string integer) :return-type :integer)

     (mus-prescaler :entry-point "mus_file_prescaler" :arguments (integer) :return-type :single-float)
     (mus-set-prescaler :entry-point "mus_file_set_prescaler" :arguments (integer single-float) :return-type :single-float :prototype t)
     #+linux (set-oss-buffers :entry-point "mus_audio_set_oss_buffers" :arguments (integer integer) :return-type :void)

     ;; the next function added to support large convolutions
     (basic-convolve :entry-point "mus_convolution" :arguments ((simple-array double-float (*)) (simple-array double-float (*)) integer) :prototype t)
     (clm-fft :entry-point "mus_fft" :arguments ((simple-array double-float (*)) (simple-array double-float (*)) integer integer) :prototype t)
     (mus-set-rand-seed :entry-point "mus_set_rand_seed" :return-type :void :arguments (integer))
     (clm-random :entry-point "mus_frandom" :return-type :double-float :prototype t :arguments (double-float))
     
     (sound-maxamp :arguments (string integer array array) :return-type :integer :entry-point "clm_sound_maxamp")
     (clm-mix :arguments (string string integer integer integer) :entry-point "clm_mix" :return-type :void)
     (clm-scale-file :arguments (string string double-float integer integer) :entry-point "clm_scale_file" :return-type :integer)
     (format->bytes :entry-point "mus_bytes_per_sample" :arguments (integer) :return-type :integer)

     (read-audio-state :entry-point "mus_audio_mixer_read" :arguments (integer integer integer array) :return-type :integer)
     (write-audio-state :entry-point "mus_audio_mixer_write" :arguments (integer integer integer array) :return-type :integer)
     (describe-audio :entry-point "mus_audio_describe" :arguments nil :return-type :void)

     (reset-audio :entry-point "mus_reset_audio_c" :arguments nil :return-type :void)
     (reset-headers :entry-point "mus_reset_headers_c" :arguments nil :return-type :void)
     (reset-io :entry-point "mus_reset_io_c" :arguments nil :return-type :void)

     (clm-close-output-1 "clm_close_output" :arguments nil :return-type :void)
     (clm-close-reverb-1 "clm_close_reverb" :arguments nil :return-type :void)     
     (clm-make-output "clm_make_output" :arguments (string integer integer integer string) :return-type :integer)
     (clm-make-reverb "clm_make_reverb" :arguments (string integer integer integer string) :return-type :integer)
     (clm-continue-output "clm_continue_output" :arguments (string) :return-type :integer)
     (clm-continue-reverb "clm_continue_reverb" :arguments (string) :return-type :integer)
     (mus-file-data-clipped-default "mus_file_data_clipped_default" :arguments nil :return-type :integer)
     (mus-file-set-data-clipped-default "mus_file_set_data_clipped_default" :arguments (integer) :return-type :integer)     
     (mus-file-prescaler-default "mus_file_prescaler_default" :arguments nil :return-type :double-float)
     (mus-file-prescaler-default "mus_file_set_prescaler_default" :arguments (double-float) :return-type :double-float)     
     (mus-srate "mus_srate" :arguments nil :return-type :double-float)
     (mus-set-srate "mus_srate" :arguments (double-float) :return-type :double-float)     
     (mus-file-buffer-size "mus_file_buffer_size" :arguments nil :return-type :integer)
     (mus-set-file-buffer-size "mus_set_file_buffer_size" :arguments (integer) :return-type :integer)     
     ))


;;; ---------------------------------------- ACL 5.n ----------------------------------------
#+(and excl acl-50)
  (progn
    ;; nearly all of these could include :call-direct t :arg-checking nil
    ;;   but my timing tests indicate that those arguments give less than a 2% speed-up (if any at all)

    #+with-clm-x (ff:def-foreign-call (init-sc "init_sc") ((version (* :char) #-acl-60 string)) :returning :void)
    
    ;; io.c
    (ff:def-foreign-call (mus-set-raw-header-defaults "mus_header_set_raw_defaults") ((srate :int) (chans :int) (format :int)) :returning :void)
    (ff:def-foreign-call (mus-set-header-type "mus_file_set_header_type") ((tfd :int) (type :int)) :returning :int)
    (ff:def-foreign-call (mus-set-data-clipped "mus_file_set_data_clipped") ((tfd :int) (type :int)) :returning :int)
    (ff:def-foreign-call (mus-set-chans "mus_file_set_chans") ((tfd :int) (type :int)) :returning :int)

    (ff:def-foreign-call (sound-chans-1 "mus_sound_chans")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-duration-1 "mus_sound_duration")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :float)
    (ff:def-foreign-call (sound-data-format-1 "mus_sound_data_format")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-data-location-1 "clm_sound_data_location")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-datum-size-1 "mus_sound_datum_size")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-header-type-1 "mus_sound_header_type")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-length-1 "clm_sound_length")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-samples-1 "clm_sound_samples")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-frames-1 "clm_sound_frames")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-srate-1 "mus_sound_srate")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    #-acl-60 (ff:def-foreign-call (sound-comment-1 "mus_sound_comment")
					 ((name (* :char) #-acl-60 string))
					 #+acl-61 :strings-convert #+acl-61 t 
					 :returning :int)
    #+acl-60 (ff:def-foreign-call (sound-comment "mus_sound_comment")
					 ((name (* :char)))
					 #+acl-61 :strings-convert #+acl-61 t 
					 :returning ((* :char)))
    #-acl-60 (ff:def-foreign-call (sound-format-name-1 "mus_data_format_name") ((form :int)) :returning :int)
    #+acl-60 (ff:def-foreign-call (sound-format-name "mus_data_format_name") ((form :int)) :returning ((* :char)))
    (ff:def-foreign-call (sound-loop-info "clm_sound_loop_info")
			 ((name (* :char) #-acl-60 string) (vals (* :int) array))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (sound-set-loop-info "mus_sound_set_full_loop_info")
			 ((name (* :char) #-acl-60 string) (vals (* :int) array))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :void)
    #-acl-60 (ff:def-foreign-call (sound-type-name-1 "mus_header_type_name") ((name :int)) :returning :int)
    #+acl-60 (ff:def-foreign-call (sound-type-name "mus_header_type_name") ((name :int)) :returning ((* :char)))
    (ff:def-foreign-call (open-sound-input "mus_sound_open_input")
			 ((name (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (open-sound-output "mus_sound_open_output")
			 ((name (* :char) #-acl-60 string) (srate :int) (chans :int) (format :int) (type :int)
			  (comment (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (reopen-sound-output "clm_sound_reopen_output")
			 ((name (* :char) #-acl-60 string) (chans :int)
			  (format :int) (type :int) (loc :int))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    (ff:def-foreign-call (close-sound-input "mus_sound_close_input") ((fd :int)) :returning :int)
    (ff:def-foreign-call (close-sound-output "clm_sound_close_output") ((fd :int) (bytes :int)) :returning :int)
    (ff:def-foreign-call (read-sound "mus_sound_read") ((fd :int) (beg :int) (end :int) (chans :int) (bufs (* :float) array)) :returning :int)
    (ff:def-foreign-call (write-sound "mus_sound_write") ((fd :int) (beg :int) (end :int) (chans :int) (bufs (* :float) array)) :returning :int)
    (ff:def-foreign-call (seek-sound-frame "clm_sound_seek_frame") ((fd :int) (frame :int)) :returning :int)
    (ff:def-foreign-call (sl-dac-1 "sl_dac")
			 ((name (* :char) #-acl-60 string) (dev :int))
			 #+acl-61 :strings-convert #+acl-61 t
			 :returning :int)
     
    (ff:def-foreign-call (mus-prescaler "mus_file_prescaler") ((fd :int)) :returning :float)
    (ff:def-foreign-call (mus-set-prescaler "mus_file_set_prescaler") ((fd :int) (val :float)) :returning :float)
    
    #-windoze (ff:def-foreign-call (clm-send-snd "clm_send_snd")
				   ((command (* :char) #-acl-60 string))
				   #+acl-61 :strings-convert #+acl-61 t 
				   :returning :int)
    #-windoze (ff:def-foreign-call (clm-start-snd "clm_start_snd") (:void) :returning :int)
    #-windoze (ff:def-foreign-call (clm-init-x "clm_init_x")
				   ((command (* :char) #-acl-60 string))
				   #+acl-61 :strings-convert #+acl-61 t 
				   :returning :void)
    #-(or acl-60 windoze) (ff:def-foreign-call (clm-receive-snd-1 "clm_receive_snd") (:void) :returning :int)
    #+(and acl-60 (not windoze)) (ff:def-foreign-call (clm-receive-snd "clm_receive_snd") (:void) :returning ((* :char)))

    (ff:def-foreign-call (initialize-sndlib "mus_sound_initialize") (:void) :returning :int)
    (ff:def-foreign-call (initialize-cmus "initialize_cmus") (:void) :returning :void)
    (ff:def-foreign-call (sound-maxamp "clm_sound_maxamp")
					     ((ifile (* :char) #-acl-60 string)
					      (chans :int)
					      #-allegro-v7.0 (arr (* :double) (simple-array double-float (*)))
					      #+allegro-v7.0 (arr (:array :double))
					      #-allegro-v7.0 (times (* :int) array)
					      #+allegro-v7.0 (times (:array :int))
					      )
					     #+acl-61 :strings-convert #+acl-61 t 
					     :returning :int)
    
    ;; audio.c
    #+linux (ff:def-foreign-call (set-oss-buffers "mus_audio_set_oss_buffers") ((num :int) (size :int)) :returning :void)
    (ff:def-foreign-call (read-audio-state "mus_audio_mixer_read")
			 ((dev :int) (field :int) (chan :int)
			  #-allegro-v7.0 (val (* :float) (simple-array single-float (*)))
			  #+allegro-v7.0 (val (:array :float))
			  )
			 :returning :int)
    (ff:def-foreign-call (write-audio-state "mus_audio_mixer_write")
			 ((dev :int) (field :int) (chan :int)
			  #-allegro-v7.0 (val (* :float) (simple-array single-float (*)))
			  #+allegro-v7.0 (val (:array :float))
			  )
			 :returning :int)
    (ff:def-foreign-call (describe-audio "mus_audio_describe") (:void) :returning :void)

    (ff:def-foreign-call (reset-audio "mus_reset_audio_c") (:void) :returning :void)
    (ff:def-foreign-call (reset-headers "mus_reset_headers_c") (:void) :returning :void)
    (ff:def-foreign-call (reset-io "mus_reset_io_c") (:void) :returning :void)

    ;; cmus.c
    #-allegro-v7.0
    (ff:def-foreign-call (basic-convolve "mus_convolution")
			 ((rl1 (* :double) (simple-array double-float (*))) (rl2 (* :double) (simple-array double-float (*))) (n :int))
			 :returning :void)
    #+allegro+v7.0
    (ff:def-foreign-call (basic-convolve "mus_convolution")
			 ((rl1 (:array :double)) (rl2 (:array :double)) (n :int))
			 :returning :void)
    #-allegro-v7.0
    (ff:def-foreign-call (clm-fft "mus_fft")
			 ((rl (* :double) (simple-array double-float (*))) (im (* :double) (simple-array double-float (*))) (n :int) (isign :int))
			 :returning :void)
    #+allegro-v7.0
    (ff:def-foreign-call (clm-fft "mus_fft")
			 ((rl (:array :double)) (im (:array :double)) (n :int) (isign :int))
			 :returning :void)
    (ff:def-foreign-call (mus-set-rand-seed "mus_set_rand_seed") ((seed :int)) :returning :void)
    (ff:def-foreign-call (clm-random "mus_frandom") ((amp :double)) :returning :double)
    #-allegro-v7.0
    (ff:def-foreign-call (array->file "clm_array2file")
			 ((file (* :char) #-acl-60 string) (data (* :double) (simple-array double-float (*))) (len :int) (srate :int) (chans :int))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :void)
    #+allegro-v7.0
    (ff:def-foreign-call (array->file "clm_array2file")
			 ((file (* :char) #-acl-60 string) (data (:array :double)) (len :int) (srate :int) (chans :int))
			 :strings-convert t 
			 :returning :void)
     (ff:def-foreign-call (clm-mix "clm_mix")
 			 ((outfile (* :char) #-acl-60 string) (infile (* :char) #-acl-60 string)
 			  (out_start :int) (out_frames :int) (in_start :int))
 			 #+acl-61 :strings-convert #+acl-61 t 
 			 :returning :void)
    (ff:def-foreign-call (clm-scale-file "clm_scale_file")
			 ((outfile (* :char) #-acl-60 string) (infile (* :char) #-acl-60 string) (scaler :double)
			  (frm :int) (hdr :int))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    #-allegro-v7.0
    (ff:def-foreign-call (file->array "clm_file2array") 
			 ((file (* :char) #-acl-60 string) (chan :int) (start :int) (samples :int)
			  (arr (* :double) (simple-array double-float (*))))
			 #+acl-61 :strings-convert #+acl-61 t 
			 :returning :int)
    #+allegro-v7.0
    (ff:def-foreign-call (file->array "clm_file2array") 
			 ((file (* :char) string) (chan :int) (start :int) (samples :int) (arr (:array :double)))
			 :strings-convert t 
			 :returning :int)

    (ff:def-foreign-call (format->bytes "mus_bytes_per_sample") ((frm :int)) :returning :int)

    (ff:def-foreign-call (clm-close-output-1 "clm_close_output") (:void) :returning :int)
    (ff:def-foreign-call (clm-close-reverb-1 "clm_close_reverb") (:void) :returning :int)
    
    (ff:def-foreign-call (clm-make-output "clm_make_output")
			 ((file (* :char) #-acl-60 string) (chans :int) (frm :int) (typ :int) (com (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 			 
			 :returning :int)
    (ff:def-foreign-call (clm-make-reverb "clm_make_reverb")
			 ((file (* :char) #-acl-60 string) (chans :int) (frm :int) (typ :int) (com (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 			 
			 :returning :int)

    (ff:def-foreign-call (clm-continue-output "clm_continue_output") ((file (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 			 
			 :returning :int)
    (ff:def-foreign-call (clm-continue-reverb "clm_continue_reverb") ((file (* :char) #-acl-60 string))
			 #+acl-61 :strings-convert #+acl-61 t 			 
			 :returning :int)

    (ff:def-foreign-call (mus-file-data-clipped-default "mus_file_data_clipped_default") (:void) :returning :int)
    (ff:def-foreign-call (mus-file-set-data-clipped-default "mus_file_set_data_clipped_default") ((clipped :int)) :returning :int)    
    (ff:def-foreign-call (mus-file-prescaler-default "mus_file_prescaler_default") (:void) :returning :double)
    (ff:def-foreign-call (mus-file-set-prescaler-default "mus_file_set_prescaler_default") ((prescaler :double)) :returning :double)    

    (ff:def-foreign-call (mus-srate "mus_srate") (:void) :returning :double)
    (ff:def-foreign-call (mus-set-srate "mus_set_srate") ((srate :double)) :returning :double)
    (ff:def-foreign-call (mus-file-buffer-size "mus_file_buffer_size") (:void) :returning :int)
    (ff:def-foreign-call (mus-set-file-buffer-size "mus_set_file_buffer_size") ((clipped :int)) :returning :int)    

   )

#+(and excl (not acl-60) (not windoze)) (defun clm-receive-snd () (let ((ptr (clm-receive-snd-1))) (if (not (zerop ptr)) (ff:char*-to-string ptr))))


;;; ---------------------------------------- CMUCL ----------------------------------------
#+cmu (use-package "ALIEN")
;#+cmu (use-package "C-CALL")
;; from cmu docs (cmucl/doc/devenv/cmu-new/node242.html)
#+cmu (defun array-data-address (array)
       "Return the physical address of where the actual data of an array is
     stored.

     ARRAY must be a specialized array type in CMU Lisp.  This means ARRAY
     must be an array of one of the following types:

                       double-float
                       single-float
                       (unsigned-byte 32)
                       (unsigned-byte 16)
                       (unsigned-byte  8)
                       (signed-byte 32)
                       (signed-byte 16)
                       (signed-byte  8)
     "
       (declare (type (or #+signed-array (array (signed-byte 8))
                          #+signed-array (array (signed-byte 16))
                          #+signed-array (array (signed-byte 32))
                          (array (unsigned-byte 8))
                          (array (unsigned-byte 16))
                          (array (unsigned-byte 32))
                          (array single-float)
                          (array double-float))
                      array)
                (optimize (speed 3) (safety 0))
                (ext:optimize-interface (safety 3)))
       ;; with-array-data will get us to the actual data.  However, because
       ;; the array could have been displaced, we need to know where the
       ;; data starts.
       (lisp::with-array-data ((data array)
                               (start)
                               (end))
         (declare (ignore end))
         ;; DATA is a specialized simple-array.  Memory is laid out like this:
         ;;
         ;;   byte offset    Value
         ;;        0         type code (should be 70 for double-float vector)
         ;;        4         4 * number of elements in vector
         ;;        8         1st element of vector
         ;;      ...         ...
         ;;
         (let ((addr (+ 8 (logandc1 7 (kernel:get-lisp-obj-address data))))
               (type-size (let ((type (array-element-type data)))
                            (cond ((or (equal type '(signed-byte 8))
                                       (equal type '(unsigned-byte 8)))
                                   1)
                                  ((or (equal type '(signed-byte 16))
                                       (equal type '(unsigned-byte 16)))
                                   2)
                                  ((or (equal type '(signed-byte 32))
                                       (equal type '(unsigned-byte 32)))
                                   4)
                                  ((equal type 'single-float)
                                   4)
                                  ((equal type 'double-float)
                                   8)
                                  (t
                                   (error "Unknown specialized array element type"))))))
           (declare (type (unsigned-byte 32) addr)
                    (optimize (speed 3) (safety 0) (ext:inhibit-warnings 3)))
           (system:int-sap (the (unsigned-byte 32)
                             (+ addr (* type-size start)))))))
#+cmu
(progn

  (defun fullname (n)
    (if (stringp n)
	(namestring (full-merge-pathnames n))
      (if (mus-input? n)
	  (mus-file-name n)
	(progn
	  (warn "~A should be a filename" n)
	  n))))
  (defun sound-chans (name) (sound-chans-1 (fullname name)))
  (defun sound-duration (name) (sound-duration-1 (fullname name)))
  (defun sound-data-format (name) (sound-data-format-1 (fullname name)))
  (defun sound-data-location (name) (sound-data-location-1 (fullname name)))
  (defun sound-datum-size (name) (sound-datum-size-1 (fullname name)))
  (defun sound-header-type (name) (sound-header-type-1 (fullname name)))
  (defun sound-length (name) (sound-length-1 (fullname name)))
  (defun sound-samples (name) (sound-samples-1 (fullname name)))
  (defun sound-frames (name) (sound-frames-1 (fullname name)))
  (defun sound-srate (name) (sound-srate-1 (fullname name)))
  (defun sound-comment (name) (sound-comment-1 (fullname name)))

    #+with-clm-x (def-alien-routine ("init_sc" init-sc) c-call:void (version c-call:c-string))
    ;; io.c
    (def-alien-routine ("mus_header_set_raw_defaults" mus-set-raw-header-defaults) c-call:void (srate c-call:int) (chans c-call:int) (format c-call:int))
    (def-alien-routine ("mus_file_set_header_type" mus-set-header-type) c-call:int (tfd c-call:int) (type c-call:int))
    (def-alien-routine ("mus_file_set_data_clipped" mus-set-data-clipped) c-call:int (tfd c-call:int) (type c-call:int))
    (def-alien-routine ("mus_file_set_chans" mus-set-chans) c-call:int (tfd c-call:int) (type c-call:int))

    (def-alien-routine ("mus_sound_chans" sound-chans-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("mus_sound_duration" sound-duration-1) c-call:float (name c-call:c-string))
    (def-alien-routine ("mus_sound_data_format" sound-data-format-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("clm_sound_data_location" sound-data-location-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("mus_sound_datum_size" sound-datum-size-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("mus_sound_header_type" sound-header-type-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("clm_sound_length" sound-length-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("clm_sound_samples" sound-samples-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("clm_sound_frames" sound-frames-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("mus_sound_srate" sound-srate-1) c-call:int (name c-call:c-string))
    (def-alien-routine ("mus_sound_comment" sound-comment-1) c-call:c-string (name c-call:c-string))
    (def-alien-routine ("mus_data_format_name" sound-format-name) c-call:c-string (form c-call:int))
    (def-alien-routine ("clm_sound_loop_info" sound-loop-info-1) c-call:int (name c-call:c-string) (vals (* c-call:int)))
    (defun sound-loop-info (name vals) (sound-loop-info-1 name (array-data-address vals)))
    (def-alien-routine ("mus_sound_set_full_loop_info" sound-set-loop-info-1) c-call:void (name c-call:c-string) (vals (* c-call:int)))
    (defun sound-set-loop-info (name vals) (sound-set-loop-info-1 name (array-data-address vals)))
    (def-alien-routine ("mus_header_type_name" sound-type-name) c-call:c-string (name c-call:int))
    (def-alien-routine ("mus_sound_open_input" open-sound-input) c-call:int (name c-call:c-string))
    (def-alien-routine ("mus_sound_open_output" open-sound-output) c-call:int
      (name c-call:c-string) (srate c-call:int) (chans c-call:int) (format c-call:int) (type c-call:int) (comment c-call:c-string))
    (def-alien-routine ("mus_sound_reopen_output" reopen-sound-output) c-call:int
      (name c-call:c-string) (chans c-call:int)  (format c-call:int) (type c-call:int) (loc c-call:int))
    (def-alien-routine ("mus_sound_close_input" close-sound-input) c-call:int (fd c-call:int))
    (def-alien-routine ("mus_sound_close_output" close-sound-output) c-call:int (fd c-call:int) (bytes c-call:int))
    (def-alien-routine ("mus_sound_read" read-sound-1) c-call:int (fd c-call:int) (beg c-call:int) (end c-call:int) (chans c-call:int) (bufs (* c-call:int)))
    (defun read-sound (fd beg end chans bufs) (read-sound-1 fd beg end chans (array-data-address bufs)))
    (def-alien-routine ("mus_sound_write" write-sound-1) c-call:int (fd c-call:int) (beg c-call:int) (end c-call:int) (chans c-call:int) (bufs (* c-call:int)))
    (defun write-sound (fd beg end chans bufs) (write-sound-1 fd beg end chans (array-data-address bufs)))
    (def-alien-routine ("mus_sound_seek_frame" seek-sound-frame) c-call:int (fd c-call:int) (frame c-call:int))
    (def-alien-routine ("sl_dac" sl-dac-1) c-call:int (name c-call:c-string) (dev c-call:int))
     
    (def-alien-routine ("mus_file_prescaler" mus-prescaler) c-call:float (fd c-call:int))
    (def-alien-routine ("mus_file_set_prescaler" mus-set-prescaler) c-call:float (fd c-call:int) (val double-float))

    #-windoze (def-alien-routine ("clm_send_snd" clm-send-snd) c-call:int (command c-call:c-string))
    #-windoze (def-alien-routine ("clm_start_snd" clm-start-snd) c-call:int)
    #-windoze (def-alien-routine ("clm_init_x" clm-init-x) c-call:void (command c-call:c-string))
    #-windoze (def-alien-routine ("clm_receive_snd" clm-receive-snd) c-call:c-string)

    (def-alien-routine ("mus_sound_initialize" initialize-sndlib) c-call:int)
    (def-alien-routine ("initialize_cmus" initialize-cmus) c-call:void)
    (def-alien-routine ("clm_sound_maxamp" sound-maxamp-1) c-call:int (ifile c-call:c-string) (chans c-call:int) (vals (* double-float)) (times (* c-call:int)))
    (defun sound-maxamp (file chans arr times) (sound-maxamp-1 file chans (array-data-address arr) (array-data-address times)))

    ;; audio.c
    #+linux (def-alien-routine ("mus_audio_set_oss_buffers" set-oss-buffers) c-call:void (num c-call:int) (size c-call:int))
    (def-alien-routine ("mus_audio_mixer_read" read-audio-state-1) c-call:int (dev c-call:int) (field c-call:int) (chan c-call:int) (val (* double-float)))
    (defun read-audio-state (dev field chan val) (read-audio-state-1 dev field chan (array-data-address val)))
    (def-alien-routine ("mus_audio_mixer_write" write-audio-state-1) c-call:int (dev c-call:int) (field c-call:int) (chan c-call:int) (val (* double-float)))
    (defun write-audio-state (dev field chan val) (write-audio-state-1 dev field chan (array-data-address val)))
    (def-alien-routine ("mus_audio_describe" describe-audio) c-call:void)
    (def-alien-routine ("mus_reset_audio_c" reset-audio) c-call:void)
    (def-alien-routine ("mus_reset_headers_c" reset-headers) c-call:void)
    (def-alien-routine ("mus_reset_io_c" reset-io) c-call:void)

    ;; cmus.c
    (def-alien-routine ("mus_convolution" basic-convolve-1) c-call:void (rl1 (* double-float)) (im1 (* double-float)) (len c-call:int))
    (defun basic-convolve (rl1 im1 len) (basic-convolve-1 (array-data-address rl1) (array-data-address im1) len))
    (def-alien-routine ("mus_fft" clm-fft-1) c-call:void (rl (* double-float)) (im (* double-float)) (len c-call:int) (isign c-call:int))
    (defun clm-fft (rl im len isign) (clm-fft-1 (array-data-address rl) (array-data-address im) len isign))
    (def-alien-routine ("mus_set_rand_seed" mus-set-rand-seed) c-call:void (seed c-call:int))
    (def-alien-routine ("mus_frandom" clm-random) c-call:float (amp double-float))
    (def-alien-routine ("clm_array2file" array->file-1) c-call:void (file c-call:c-string) (data (* double-float)) (len c-call:int) (srate c-call:int) (chans c-call:int))
    (defun array->file (file data len srate chans) (array->file-1 file (array-data-address data) len srate chans))
    (def-alien-routine ("clm_mix" clm-mix) 
 	c-call:void
 	(outfile c-call:c-string) (infile c-call:c-string)  (out_start c-call:int) (out_frames c-call:int) (in_start c-call:int))
    (def-alien-routine ("clm_scale_file" clm-scale-file) c-call:int (outfile c-call:c-string) (infile c-call:c-string) (scaler double-float) (frm c-call:int) (hdr c-call:int))
    (def-alien-routine ("clm_file2array" file->array-1) 
	c-call:int 
	(file c-call:c-string) (chan c-call:int) (start c-call:int) (samples c-call:int)  (arr (* double-float)))
    (defun file->array (file chan start samples arr) (file->array-1 file chan start samples (array-data-address arr)))
    (def-alien-routine ("mus_bytes_per_sample" format->bytes) c-call:int (frm c-call:int))

    (def-alien-routine ("clm_close_output" clm-close-output-1) c-call:int)
    (def-alien-routine ("clm_close_reverb" clm-close-reverb-1) c-call:int)
    (def-alien-routine ("clm_make_output" clm-make-output) c-call:int
      (file c-call:c-string) (chans c-call:int) (frm c-call:int) (typ c-call:int) (com c-call:c-string))
    (def-alien-routine ("clm_make_reverb" clm-make-reverb) c-call:int
      (file c-call:c-string) (chans c-call:int) (frm c-call:int) (typ c-call:int) (com c-call:c-string))
    (def-alien-routine ("clm_continue_output" clm-continue-output) c-call:int (file c-call:c-string))
    (def-alien-routine ("clm_continue_reverb" clm-continue-reverb) c-call:int (file c-call:c-string))
    (def-alien-routine ("mus_file_data_clipped_default" mus-file-data-clipped-default) c-call:int)
    (def-alien-routine ("mus_file_set_data_clipped_default" mus-file-set-data-clipped-default) c-call:int (val c-call:int))
    (def-alien-routine ("mus_file_prescaler_default" mus-file-prescaler-default) c-call:float)
    (def-alien-routine ("mus_file_set_prescaler_default" mus-file-set-prescaler-default) double-float (val double-float))
    (def-alien-routine ("mus_srate" mus-srate) double-float)
    (def-alien-routine ("mus_set_srate" mus-set-srate) double-float (val double-float))
    (def-alien-routine ("mus_file_buffer_size" mus-file-buffer-size) c-call:int)
    (def-alien-routine ("mus_set_file_buffer_size" mus-set-file-buffer-size) c-call:int (val c-call:int))
    )

;for all: (declaim (inline lisp-name))


;;; ---------------------------------------- SBCL ----------------------------------------
;#+sbcl (use-package "SB-ALIEN")

;; from contrib/sb-grovel/array-data.lisp
#+sbcl (defun array-data-address (array)
	 "Return the physical address of where the actual data of an array is
stored.

ARRAY must be a specialized array type - an array of one of these types:

                  double-float
                  single-float
                  (unsigned-byte 32)
                  (unsigned-byte 16)
                  (unsigned-byte  8)
                  (signed-byte 32)
                  (signed-byte 16)
                  (signed-byte  8)
"
  (declare (type (or (array (signed-byte 8))
		     (array base-char)
		     simple-base-string
                     (array (signed-byte 16))
                     (array (signed-byte 32))
                     (array (unsigned-byte 8))
                     (array (unsigned-byte 16))
                     (array (unsigned-byte 32))
                     (array single-float)
                     (array double-float))
                 array)
           (optimize (speed 0) (debug 3) (safety 3)))
  ;; with-array-data will get us to the actual data.  However, because
  ;; the array could have been displaced, we need to know where the
  ;; data starts.

  (let* ((type (car (multiple-value-list (array-element-type array))))
	 (type-size
	  (cond ((or (equal type '(signed-byte 8))
		     (equal type 'cl::base-char)
		     (equal type '(unsigned-byte 8)))
		 1)
		((or (equal type '(signed-byte 16))
		     (equal type '(unsigned-byte 16)))
		 2)
		((or (equal type '(signed-byte 32))
		     (equal type '(unsigned-byte 32)))
		 4)
		((equal type 'single-float)
		 4)
		((equal type 'double-float)
		 8)
		(t (error "Unknown specialized array element type")))))
    (sb-kernel::with-array-data ((data array)
		      (start)
		      (end))
      (declare (ignore end))
      ;; DATA is a specialized simple-array.  Memory is laid out like this:
      ;;
      ;;   byte offset    Value
      ;;        0         type code (e.g. 70 for double-float vector)
      ;;        4         FIXNUMIZE(number of elements in vector)
      ;;        8         1st element of vector
      ;;      ...         ...
      ;;
      (let* ((addr (+ 8 (logandc1 7 (sb-kernel:get-lisp-obj-address data)))))
	(declare (type (unsigned-byte 32) addr)
		 (optimize (speed 3) (safety 0)))
	(sb-sys:int-sap (the (unsigned-byte 32)
			  (+ addr (* type-size start))))))))


#+sbcl
(progn
    (defun fullname (n)
    (if (stringp n)
	(namestring (full-merge-pathnames n))
      (if (mus-input? n)
	  (mus-file-name n)
	(progn
	  (warn "~A should be a filename" n)
	  n))))
  (defun sound-chans (name) (sound-chans-1 (fullname name)))
  (defun sound-duration (name) (sound-duration-1 (fullname name)))
  (defun sound-data-format (name) (sound-data-format-1 (fullname name)))
  (defun sound-data-location (name) (sound-data-location-1 (fullname name)))
  (defun sound-datum-size (name) (sound-datum-size-1 (fullname name)))
  (defun sound-header-type (name) (sound-header-type-1 (fullname name)))
  (defun sound-length (name) (sound-length-1 (fullname name)))
  (defun sound-samples (name) (sound-samples-1 (fullname name)))
  (defun sound-frames (name) (sound-frames-1 (fullname name)))
  (defun sound-srate (name) (sound-srate-1 (fullname name)))
  (defun sound-comment (name) (sound-comment-1 (fullname name)))
  
    #+with-clm-x (sb-alien:define-alien-routine ("init_sc" init-sc) sb-alien:void (version sb-alien:c-string))
    ;; io.c
    (sb-alien:define-alien-routine ("mus_header_set_raw_defaults" mus-set-raw-header-defaults) sb-alien:void
				   (srate sb-alien:int) (chans sb-alien:int) (format sb-alien:int))
    (sb-alien:define-alien-routine ("mus_file_set_header_type" mus-set-header-type) sb-alien:int (tfd sb-alien:int) (type sb-alien:int))
    (sb-alien:define-alien-routine ("mus_file_set_data_clipped" mus-set-data-clipped) sb-alien:int (tfd sb-alien:int) (type sb-alien:int))
    (sb-alien:define-alien-routine ("mus_file_set_chans" mus-set-chans) sb-alien:int (tfd sb-alien:int) (type sb-alien:int))
    (sb-alien:define-alien-routine ("mus_sound_chans" sound-chans-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_sound_duration" sound-duration-1) sb-alien:float (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_sound_data_format" sound-data-format-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_sound_data_location" sound-data-location-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_sound_datum_size" sound-datum-size-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_sound_header_type" sound-header-type-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_sound_length" sound-length-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_sound_samples" sound-samples-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_sound_frames" sound-frames-1) sb-alien:int (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_sound_srate" sound-srate-1) sb-alien:int (name sb-alien:c-string))

    (sb-alien:define-alien-routine ("mus_sound_comment" sound-comment-1) sb-alien:c-string (name sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_data_format_name" sound-format-name) sb-alien:c-string (form sb-alien:int))
    (sb-alien:define-alien-routine ("clm_sound_loop_info" sound-loop-info-1) sb-alien:int (name sb-alien:c-string) (vals (* sb-alien:int)))
    (defun sound-loop-info (name vals) (sound-loop-info-1 name (array-data-address vals)))

    (sb-alien:define-alien-routine ("mus_sound_set_full_loop_info" sound-set-loop-info-1) sb-alien:void
				   (name sb-alien:c-string) (vals (* sb-alien:int)))
    (defun sound-set-loop-info (name vals) (sound-set-loop-info-1 name (array-data-address vals)))
    (sb-alien:define-alien-routine ("mus_header_type_name" sound-type-name) sb-alien:c-string (name sb-alien:int))
    (sb-alien:define-alien-routine ("mus_sound_open_input" open-sound-input) sb-alien:int (name sb-alien:c-string))

    (sb-alien:define-alien-routine ("mus_sound_open_output" open-sound-output) sb-alien:int
				   (name sb-alien:c-string) (srate sb-alien:int) (chans sb-alien:int) (format sb-alien:int)
				   (type sb-alien:int) (comment sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_sound_reopen_output" reopen-sound-output) sb-alien:int
				   (name sb-alien:c-string) (chans sb-alien:int)  (format sb-alien:int) (type sb-alien:int) (loc sb-alien:int))
    (sb-alien:define-alien-routine ("mus_sound_close_input" close-sound-input) sb-alien:int (fd sb-alien:int))
    (sb-alien:define-alien-routine ("mus_sound_close_output" close-sound-output) sb-alien:int (fd sb-alien:int) (bytes sb-alien:int))

    (sb-alien:define-alien-routine ("mus_sound_read" read-sound-1) sb-alien:int
				   (fd sb-alien:int) (beg sb-alien:int) (end sb-alien:int) (chans sb-alien:int) (bufs (* sb-alien:int)))
    (defun read-sound (fd beg end chans bufs) (read-sound-1 fd beg end chans (array-data-address bufs)))
    (sb-alien:define-alien-routine ("mus_sound_write" write-sound-1) sb-alien:int
				   (fd sb-alien:int) (beg sb-alien:int) (end sb-alien:int) (chans sb-alien:int) (bufs (* sb-alien:int)))
    (defun write-sound (fd beg end chans bufs) (write-sound-1 fd beg end chans (array-data-address bufs)))
    (sb-alien:define-alien-routine ("mus_sound_seek_frame" seek-sound-frame) sb-alien:int (fd sb-alien:int) (frame sb-alien:int))
    (sb-alien:define-alien-routine ("sl_dac" sl-dac-1) sb-alien:int (name sb-alien:c-string) (dev sb-alien:int))
     
    (sb-alien:define-alien-routine ("mus_file_prescaler" mus-prescaler) sb-alien:float (fd sb-alien:int))
    (sb-alien:define-alien-routine ("mus_file_set_prescaler" mus-set-prescaler) sb-alien:float (fd sb-alien:int) (val double-float))
    
    #-windoze (sb-alien:define-alien-routine ("clm_send_snd" clm-send-snd) sb-alien:int (command sb-alien:c-string))
    #-windoze (sb-alien:define-alien-routine ("clm_start_snd" clm-start-snd) sb-alien:int)
    #-windoze (sb-alien:define-alien-routine ("clm_init_x" clm-init-x) sb-alien:void (command sb-alien:c-string))
    #-windoze (sb-alien:define-alien-routine ("clm_receive_snd" clm-receive-snd) sb-alien:c-string)

    (sb-alien:define-alien-routine ("mus_sound_initialize" initialize-sndlib) sb-alien:int)
    (sb-alien:define-alien-routine ("initialize_cmus" initialize-cmus) sb-alien:void)
    (sb-alien:define-alien-routine ("clm_sound_maxamp" sound-maxamp-1) sb-alien:int
				   (ifile sb-alien:c-string) (chans sb-alien:int) (vals (* double-float)) (times (* sb-alien:int)))
    (defun sound-maxamp (file chans arr times) (sound-maxamp-1 file chans (array-data-address arr) (array-data-address times)))

    ;; audio.c
    #+linux (sb-alien:define-alien-routine ("mus_audio_set_oss_buffers" set-oss-buffers) sb-alien:void (num sb-alien:int) (size sb-alien:int))
    (sb-alien:define-alien-routine ("mus_audio_mixer_read" read-audio-state-1) sb-alien:int
				   (dev sb-alien:int) (field sb-alien:int) (chan sb-alien:int) (val (* double-float)))
    (defun read-audio-state (dev field chan val) (read-audio-state-1 dev field chan (array-data-address val)))
    (sb-alien:define-alien-routine ("mus_audio_mixer_write" write-audio-state-1) sb-alien:int
				   (dev sb-alien:int) (field sb-alien:int) (chan sb-alien:int) (val (* double-float)))
    (defun write-audio-state (dev field chan val) (write-audio-state-1 dev field chan (array-data-address val)))
    (sb-alien:define-alien-routine ("mus_audio_describe" describe-audio) sb-alien:void)
    (sb-alien:define-alien-routine ("mus_reset_audio_c" reset-audio) sb-alien:void)
    (sb-alien:define-alien-routine ("mus_reset_headers_c" reset-headers) sb-alien:void)
    (sb-alien:define-alien-routine ("mus_reset_io_c" reset-io) sb-alien:void)
    
    ;; cmus.c
    (sb-alien:define-alien-routine ("mus_convolution" basic-convolve-1) sb-alien:void
				   (rl1 (* double-float)) (im1 (* double-float)) (len sb-alien:int))
    (defun basic-convolve (rl1 im1 len) (basic-convolve-1 (array-data-address rl1) (array-data-address im1) len))
    (sb-alien:define-alien-routine ("mus_fft" clm-fft-1) sb-alien:void
				   (rl (* double-float)) (im (* double-float)) (len sb-alien:int) (isign sb-alien:int))
    (defun clm-fft (rl im len isign) (clm-fft-1 (array-data-address rl) (array-data-address im) len isign))
    (sb-alien:define-alien-routine ("mus_set_rand_seed" mus-set-rand-seed) sb-alien:void (seed sb-alien:int))
    (sb-alien:define-alien-routine ("mus_frandom" clm-random) sb-alien:float (amp double-float))
    (sb-alien:define-alien-routine ("clm_array2file" array->file-1) sb-alien:void
				   (file sb-alien:c-string) (data (* double-float)) (len sb-alien:int) (srate sb-alien:int) (chans sb-alien:int))
    (defun array->file (file data len srate chans) (array->file-1 file (array-data-address data) len srate chans))
    (sb-alien:define-alien-routine ("clm_mix" clm-mix) 
	sb-alien:void
 	(outfile sb-alien:c-string) (infile sb-alien:c-string)  (out_start sb-alien:int) (out_frames sb-alien:int) (in_start sb-alien:int))
    (sb-alien:define-alien-routine ("clm_scale_file" clm-scale-file) sb-alien:int
				   (outfile sb-alien:c-string) (infile sb-alien:c-string) (scaler double-float) (frm sb-alien:int) (hdr sb-alien:int))
    (sb-alien:define-alien-routine ("clm_file2array" file->array-1) 
	sb-alien:int 
	(file sb-alien:c-string) (chan sb-alien:int) (start sb-alien:int) (samples sb-alien:int)  (arr (* double-float)))
    (defun file->array (file chan start samples arr) (file->array-1 file chan start samples (array-data-address arr)))
    (sb-alien:define-alien-routine ("mus_bytes_per_sample" format->bytes) sb-alien:int (frm sb-alien:int))

    (sb-alien:define-alien-routine ("clm_close_output" clm-close-output-1) sb-alien:int)
    (sb-alien:define-alien-routine ("clm_close_reverb" clm-close-reverb-1) sb-alien:int)
    (sb-alien:define-alien-routine ("clm_make_output" clm-make-output) sb-alien:int
      (file sb-alien:c-string) (chans sb-alien:int) (frm sb-alien:int) (typ sb-alien:int) (com sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_make_reverb" clm-make-reverb) sb-alien:int
      (file sb-alien:c-string) (chans sb-alien:int) (frm sb-alien:int) (typ sb-alien:int) (com sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_continue_output" clm-continue-output) sb-alien:int (file sb-alien:c-string))
    (sb-alien:define-alien-routine ("clm_continue_reverb" clm-continue-reverb) sb-alien:int (file sb-alien:c-string))
    (sb-alien:define-alien-routine ("mus_file_data_clipped_default" mus-file-data-clipped-default) sb-alien:int)
    (sb-alien:define-alien-routine ("mus_file_set_data_clipped_default" mus-file-set-data-clipped-default) sb-alien:int (val sb-alien:int))
    (sb-alien:define-alien-routine ("mus_file_prescaler_default" mus-file-prescaler-default) sb-alien:double-float)
    (sb-alien:define-alien-routine ("mus_file_set_prescaler_default" mus-file-set-prescaler-default) sb-alien:double-float (val sb-alien:double-float))
    (sb-alien:define-alien-routine ("mus_srate" mus-srate) sb-alien:double-float)
    (sb-alien:define-alien-routine ("mus_set_srate" mus-set-srate) sb-alien:double-float (val sb-alien:double-float))
    (sb-alien:define-alien-routine ("mus_file_buffer_size" mus-file-buffer-size) sb-alien:int)
    (sb-alien:define-alien-routine ("mus_set_file_buffer_size" mus-set-file-buffer-size) sb-alien:int (val sb-alien:int))
    
)


;;; ---------------------------------------- OPENMCL ----------------------------------------
#+openmcl
(progn
  (defun fullname (n)
    (if (or (pathnamep n) (stringp n))
	n
      (if (mus-input? n)
	  (mus-file-name n)
        (progn
          (warn "~A should be a filename" n)
          n))))

  (defmacro with-c-filename ((p f) &body body)
    `(ccl:with-cstrs ((,p (namestring (translate-logical-pathname ,f))))
		     ,@body))

  (defun heap-float* (hfarray &aux (p (ccl:%null-ptr)))
    (ccl::%vect-data-to-macptr hfarray p)
    p)

  (defun heap-int* (hiarray &aux (p (ccl:%null-ptr)))
    (ccl::%vect-data-to-macptr hiarray p)
    p)

  (defun get-cstring (str)
    (when (not (ccl:%null-ptr-p str))
      (%get-cstring str)))

  (defun sound-chans (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_chans" :address f :signed)))
  (defun sound-data-format (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_data_format" :address f :signed)))
  (defun sound-data-location (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_data_location" :address f (:signed 64))))
  (defun sound-datum-size (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_datum_size" :address f :signed)))
  (defun sound-header-type (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_header_type" :address f :signed)))
  (defun sound-length (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_length" :address f (:signed 64))))
  (defun sound-samples (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_samples" :address f (:signed 64))))
  (defun sound-frames (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_frames" :address f (:signed 64))))
  (defun sound-srate (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_srate" :address f :signed)))
  (defun sound-duration (file) (with-c-filename (f (fullname file)) (ccl::external-call "_mus_sound_duration" :address f :single-float)))
  (defun sound-comment (file) (with-c-filename (f (fullname file)) (get-cstring (ccl::external-call "_mus_sound_comment" :address f :address))))
  (defun mus-set-raw-header-defaults (srate chans format) ; mixed-up name is correct!
    (ccl::external-call "_mus_header_set_raw_defaults" :signed srate :signed chans :signed format :void))
  (defun mus-set-header-type (tfd type) (ccl::external-call "_mus_file_set_header_type" :signed tfd :signed type :signed))
  (defun mus-set-data-clipped (tfd type) (ccl::external-call "_mus_file_set_data_clipped" :signed tfd :signed type :signed))
  (defun mus-set-chans (tfd type) (ccl::external-call "_mus_file_set_chans" :signed tfd :signed type :signed))
  (defun sound-format-name (frm) (get-cstring (ccl::external-call "_mus_data_format_name" :signed frm :address)))
  (defun sound-type-name (typ) (get-cstring (ccl::external-call "_mus_header_type_name" :signed typ :address)))
  (defun mus-sound-open-input (file) (ccl:with-cstrs ((f file)) (ccl::external-call "_mus_sound_open_input" :address f :signed)))
  (defun close-sound-input (fd) (ccl::external-call "_mus_sound_close_input" :signed fd :signed))
  (defun close-sound-output (fd bytes) (ccl::external-call "_mus_sound_close_output" :signed fd (:signed 64) bytes :signed))
  (defun seek-sound-frame (fd frame) (ccl::external-call "_mus_sound_seek_frame" :signed fd (:signed 64) frame (:signed 64)))  
  (defun reset-audio () (ccl::external-call "_mus_reset_audio_c" :void))
  (defun reset-headers () (ccl::external-call "_mus_reset_headers_c" :void))
  (defun reset-io () (ccl::external-call "_mus_reset_io_c" :void))
  (defun initialize-cmus () (ccl::external-call "_initialize_cmus" :void))
  (defun describe-audio () (ccl::external-call "_describe_audio" :void))
  (defun initialize-sndlib () (ccl::external-call "_mus_sound_initialize" :signed))
  (defun format->bytes (frm) (ccl::external-call "_mus_bytes_per_sample" :signed frm :signed))
  (defun mus-set-rand-seed (seed) (ccl::external-call "_mus_set_rand_seed" :signed seed :void))
  (defun sound-loop-info (name vals)
    (ccl:with-cstrs ((f name))
      (ccl::external-call "_clm_sound_loop_info" :address f :address (heap-int* vals))))
  (defun sound-set-loop-info (name vals)
    (ccl:with-cstrs ((f name))
      (ccl::external-call "_mus_sound_set_full_loop_info" :address f :address (heap-int* vals))))
  (defun open-sound-output (name srate chans format type comment)
    (ccl:with-cstrs ((f name) (c comment))
      (ccl::external-call "_mus_sound_open_output" :address f :signed srate :signed chans :signed format :signed type :address c :signed)))
  (defun reopen-sound-output (name chans format type loc)
    (ccl:with-cstrs ((f name))
      (ccl::external-call "_mus_sound_reopen_output" :address f :signed chans :signed format :signed type (:signed 64) loc :signed)))
  (defun open-sound-input (file)
    (ccl:with-cstrs ((f file))
      (ccl::external-call "_mus_sound_open_input" :address f :signed)))
  (defun read-sound (fd beg end chans bufs)
    (ccl::external-call "_mus_sound_read" :signed fd :signed beg :signed end :signed chans :address (heap-int* bufs) :signed))
  (defun write-sound (fd beg end chans bufs)
    (ccl::external-call "_mus_sound_write" :signed fd :signed beg :signed end :signed chans :address (heap-int* bufs) :signed))
  (defun sl-dac-1 (name dev)
    (warn "sl-dac doesn't work in openmcl"))
  
  (defun sound-maxamp (file chans arr times)
    (ccl:with-cstrs ((f file))
      (ccl::with-foreign-double-float-array (p arr)
        (ccl::external-call "_clm_sound_maxamp" :address f :signed chans :address p :address (heap-int* times) (:signed 64)))))
  (defun clm-random (amp) (ccl::external-call "_mus_frandom" :double-float amp :double-float))
  (defun array->file (file data len srate chans)
    (ccl:with-cstrs ((f file))
      (ccl::with-foreign-double-float-array (p data)
        (ccl::external-call "_clm_array2file" :address f :address p :signed len :signed srate :signed chans :void))))
  (defun clm-scale-file (ofile ifile scaler frm hdr)
    (ccl:with-cstrs ((f1 ofile) (f2 ifile))
      (ccl::external-call "_clm_scale_file" :address f1 :address f2 :double-float scaler :signed frm :signed hdr :signed)))
  (defun mus-prescaler (fd) (ccl::external-call "_mus_file_prescaler" :signed fd :single-float))
  (defun mus-set-prescaler (fd val) (ccl::external-call "_mus_file_set_prescaler" :signed fd :single-float val :single-float))
  (defun read-audio-state (dev field chan val)
    (ccl::external-call "_mus_audio_mixer_read" :signed dev :signed field :signed chan :address (heap-float* val) :signed))
  (defun write-audio-state (dev field chan val)
    (ccl::external-call "_mus_audio_mixer_write" :signed dev :signed field :signed chan :address (heap-float* val) :signed))
  (defun clm-fft (rl im len isign)
    (ccl::with-foreign-double-float-array-to-c-and-lisp (arl rl)
      (ccl::with-foreign-double-float-array-to-c-and-lisp (aim im)
        (ccl::external-call "_mus_fft" :address arl :address aim :signed len :signed isign :void))))
  (defun basic-convolve (rl im len)
    (ccl::with-foreign-double-float-array-to-c-and-lisp (arl rl)
      (ccl::with-foreign-double-float-array-to-c-and-lisp (aim im)
        (ccl::external-call "_mus_convolution" :address arl :address aim :signed len :void))))
  (defun file->array (file chan start samples arr)
    (ccl:with-cstrs ((f file))
      (ccl::with-foreign-double-float-array-to-lisp (p arr)
        (ccl::external-call "_clm_file2array" :address f :signed chan :signed start :signed samples :address p :signed))))
  (defun clm-mix (ofile ifile ostart oframes istart)
     (ccl:with-cstrs ((f1 ofile) (f2 ifile))
       (ccl::external-call "_clm_mix"
 			 :address f1 :address f2 :signed ostart :signed oframes :signed istart
 			 :void)))

  ;; these courtesy of Todd Ingalls
  #+X11
  (defun clm-send-snd (command)
    (ccl:with-cstrs ((cmd command))
      (ccl::external-call "_clm_send_snd" :address cmd :signed)))

  #+X11
  (defun clm-start-snd () (ccl::external-call "_clm_start_snd" :signed))

  #+X11
  (defun clm-init-x (command)
    (ccl:with-cstrs ((cmd command)) (ccl::external-call "_clm_init_x" :address cmd :void)))

  #+X11
  (defun clm-receive-snd ()
    (ccl:%get-cstring
      (ccl::external-call "_clm_receive_snd" :address)))

  (defun clm-close-output-1 () (ccl::external-call "_clm_close_output" :signed))
  (defun clm-close-reverb-1 () (ccl::external-call "_clm_close_reverb" :signed))
  (defun clm-make-output (file chans frm typ com)
    (ccl:with-cstrs ((f1 file)
		     (c1 com))
      (ccl::external-call "_clm_make_output"
			 :address f1 :signed chans :signed frm :signed typ :address c1
			 :signed)))
  (defun clm-make-reverb (file chans frm typ com)
    (ccl:with-cstrs ((f1 file)
		     (c1 com))
      (ccl::external-call "_clm_make_reverb"
			 :address f1 :signed chans :signed frm :signed typ :address c1
			 :signed)))
  (defun clm-continue-output (file)
    (ccl:with-cstrs ((f1 file)) (ccl::external-call "_clm_continue_output" :address f1 :signed)))    
  (defun clm-continue-reverb (file)
    (ccl:with-cstrs ((f1 file)) (ccl::external-call "_clm_continue_reverb" :address f1 :signed)))    
  (defun mus-file-data-clipped-default () (ccl::external-call "_mus_file_data_clipped_default" :signed))
  (defun mus-file-set-data-clipped-default (val) (ccl::external-call "_mus_file_set_data_clipped_default" :signed val :signed))
  (defun mus-file-prescaler-default () (ccl::external-call "_mus_file_prescaler_default" :double-float))
  (defun mus-file-set-prescaler-default (val) (ccl::external-call "_mus_file_set_prescaler_default" :double-float val :double-float))
  (defun mus-srate () (ccl::external-call "_mus_srate" :double-float))
  (defun mus-set-srate (val) (ccl::external-call "_mus_set_srate" :double-float val :double-float))
  (defun mus-file-buffer-size () (ccl::external-call "_mus_file_buffer_size" :signed))
  (defun mus-set-file-buffer-size (val) (ccl::external-call "_mus_set_file_buffer_size" :signed val :signed))
)

